#lang racket
;; Implementation of EAM's "alt-exp" syntax for reduced parens and infix that is
;; completely analogous to s-exp.  The lexer/parser is based on the example R5RS
;; parser.
;; 
;; TODO: NOTE: Can't use the provided yacc/LALR parser tool, since it has no place
;; for semantic predicates.  I need to be able to check indentation at the right
;; place.  I thought I could make a hack where the parser tells the lexer via a 
;; flag that an expr has ended and that indent should be checked on the next token
;; but that won't work when parsing an expr inside some colon-group that may span
;; multiple lines and have whatever kind of indentation but should not be closing
;; the colon group, eg.
;;    let: :[a
;;     1
;;        ] ;these two tokens should not close the colon-group!
;;          [b 2]
;;         +(a b)
;; therefore I have to make a hand-crafted LL(1) parser so I can manually check
;; the predicates before each expr that might be in the group.  Basically, the 
;; parser is the only one that knows when checking must be done, and the provided
;; parser tools do not allow actions to occur inside of a parse to pass the 
;; context info back to the lexer, nor is there direct support for predicates.
;;
;; See the comment above the parser for grammar and indentation rules.
;; 
;; Changelog:
;;   - added lambda symbol as synonymn for an id "lambda"
;;   - added #' for syntax templates
;;   - 10/28/10 changed {} to mean infix, but for backwards compatibility I keep (| |)
;;     so for lambda in application position I now use _( ) which I really didn't use
;;     before because [] does the same thing.
;;
;; Known Issues:
;;   - The last line in a program must end in \n or the parse fails (should accept eof...)
;;     Notes:
;;     checkIndent():
;;       Loop until next token isn't a possible start of an alt exp or indentation is done
;;       hmm, EOF shouldn't be in first(alt-exp), so why doesn't it stop right on EOF??
;;     
;;     I also added a rule to indent-done? to return true if the next token is EOF.  
;;     This should never apply, because the (member first(alt-exp)) is checked first in the (or ...) clause.
;;     
;;     Need to experiment with this behavior.
;;     Does it fail in non-indent groups or indent groups or both?
;;     Do comments at the end affect it? etc.
;;   - An application using CP_COLON fails to parse if the OP and CP_COLON aren't on the same line.
;;   - IDs starting with a number or - do not correctly parse as ids.
;;

(require parser-tools/lex
         (prefix-in : parser-tools/lex-sre)
         parser-tools/yacc
         syntax/readerr)

; --------------------------------------------------------------
; Setup Tokens
; --------------------------------------------------------------
(define-tokens data (DATUM ID ID_COLON))
(define-empty-tokens delim (OP CP _OP SOP SCP COP CCP CPCOLON COLON INOP INCP
                               HASHOP QUOTE QUASIQUOTE UNQUOTE UNQUOTE-SPLICING DOT EOF))


; --------------------------------------------------------------
; Lexer
; --------------------------------------------------------------
(define alt-exp-lexer
  (lexer-src-pos
   
   ;; Skip comments, without accumulating extra position information
   [(:or scheme-whitespace comment) (return-without-pos (alt-exp-lexer input-port))]
   [(:: #\# #\|) 
    (begin 
      (get-comment-token input-port) ;consume rest of multi-line comment
      (return-without-pos (alt-exp-lexer input-port)))] ;then get next token
   
   ; Constants
   ["#t" (token-DATUM #t)]
   ["#f" (token-DATUM #f)]
   [(:: "#\\" any-char) (token-DATUM (caddr (string->list lexeme)))]
   ["#\\space" (token-DATUM #\space)]
   ["#\\newline" (token-DATUM #\newline)]
   [#\" (token-DATUM (list->string (get-string-token input-port)))]
   
   ; Keyword
   [(:: #\# #\: (:or (:: initial (:* subsequent)) "+" "-" "...")) (token-DATUM (string->keyword 
                                                                                (substring lexeme 2
                                                                                           (string-length lexeme))))]
   ; ID / ID_COLON
   [(:: (:or (:: initial (:* subsequent)) "+" "-" "...") #\:) (token-ID_COLON (string->symbol 
                                                                               (substring lexeme 0 
                                                                                          (- (string-length lexeme) 1))))]
   [(:or (:: initial (:* subsequent)) "+" "-" "...") (token-ID (string->symbol lexeme))]
   
   [(:: #\λ #\:) (token-ID_COLON (string->symbol "lambda"))]
   [#\λ (token-ID (string->symbol "lambda"))]
   ;ID that starts with sign/digit --> put after number parsing, require ws/paren --> nix that ...
   ;[(:: (:* (:or "-" "+" digit)) initial (:* subsequent) #\:) (token-ID_COLON (string->symbol 
   ;                                                                (substring lexeme 0 
   ;                                                                           (- (string-length lexeme) 1))))]
   ;[(:: (:* (:or "-" "+" digit)) initial (:* subsequent)) (token-ID (string->symbol lexeme))]
   ;ID that starts with sign/digit --> this fails on 1/2
   ;[(:: (:or num10 "-" "+") initial (:* subsequent) #\:) (token-ID_COLON (string->symbol 
   ;                                                                (substring lexeme 0 
   ;                                                                           (- (string-length lexeme) 1))))]
   ;[(:: (:or num10 "-" "+") initial (:* subsequent)) (token-ID (string->symbol lexeme))]
   
   ;Arbitrary identifiers in the form #"id:goes ,&% here"
   ;TODO: let colon followed by idchar be an idchar, so exn:fail:syntax is an id. (Optional)
   [(:: #\# #\") 
    ;First, read in the string so we know the identifier
    (let ([id (string->symbol (list->string (get-string-token input-port)))])
      ;But now we check if it's an ID or ID_COLON token!
      (if (equal? (peek-char input-port) #\:)
          (begin
            (read-char input-port) ;consume the colon
            (token-ID_COLON id))
          (token-ID id)))]
   
   ; Parens, etc.
   ; TODO: attach 'paren-shape property #\[ or #\{
   ;  see Reference Racket 11.7 "Syntax Object Properties"
   ["_(" '_OP]
   ["):" 'CPCOLON]
   ["(|" 'INOP]
   ["|)" 'INCP]
   ["{" 'INOP]
   ["}" 'INCP]
   [#\( 'OP]
   [#\) 'CP]
   [#\[ 'SOP]
   [#\] 'SCP]
   ;[#\{ 'COP]
   ;[#\} 'CCP]
   [#\: 'COLON]
   
   ; Other R5Rs/Racket scheme symbols
   ["#(" 'HASHOP]
   ["#'" 'SYNTAX-QUOTE]
   [num2 (token-DATUM (string->number lexeme 2))]
   [num8 (token-DATUM (string->number lexeme 8))]
   [num10 (token-DATUM (string->number lexeme 10))]
   [num16 (token-DATUM (string->number lexeme 16))]
   ["'" 'QUOTE]
   ["`" 'QUASIQUOTE]
   ["," 'UNQUOTE]
   [",@" 'UNQUOTE-SPLICING]
   ["." 'DOT]
   [(eof) 'EOF]))

(define get-string-token
  (lexer
   [(:~ #\" #\\) (cons (car (string->list lexeme))
                       (get-string-token input-port))]
   [(:: #\\ #\\) (cons #\\ (get-string-token input-port))]
   [(:: #\\ #\") (cons #\" (get-string-token input-port))]
   [(:: #\\ #\n) (cons #\newline (get-string-token input-port))]
   [#\" null]))

(define get-comment-token
  (lexer
   [(:~ #\|) (get-comment-token input-port)]
   [(:: #\| (:~ #\#)) (get-comment-token input-port)]
   [(:: #\| #\#) null]))


(define-lex-abbrevs
  [letter (:or (:/ "a" "z") (:/ #\A #\Z))]
  [digit (:/ #\0 #\9)]
  [scheme-whitespace (:or #\newline #\return #\tab #\space #\vtab)]
  [initial (:or letter (char-set "!$%&*/<=>?^_~@"))]
  [subsequent (:or initial digit (char-set "+-.@"))]
  [comment (:: #\; (:* (:~ #\newline)) #\newline)] ;TODO: make sure nl is optional or can be EOF (but don't eat the EOF)
  
  
  ;; See ${PLTHOME}/collects/syntax-color/scheme-lexer.ss for an example of
  ;; using regexp macros to avoid the cut and paste.
  ;   [numR (:: prefixR complexR)]
  ;   [complexR (:or realR
  ;                (:: realR "@" realR)
  ;                (:: realR "+" urealR "i")
  ;                (:: realR "-" urealR "i")
  ;                (:: realR "+i")
  ;                (:: realR "-i")
  ;                (:: "+" urealR "i")
  ;                (:: "-" urealR "i")
  ;                (:: "+i")
  ;                (:: "-i"))]
  ;   [realR (:: sign urealR)]
  ;   [urealR (:or uintegerR (:: uintegerR "/" uintegerR) decimalR)]
  ;   [uintegerR (:: (:+ digitR) (:* #\#))]
  ;   [prefixR (:or (:: radixR exactness)
  ;               (:: exactness radixR))]
  
  [num2 (:: prefix2 complex2)]
  [complex2 (:or real2
                 (:: real2 "@" real2)
                 (:: real2 "+" ureal2 "i")
                 (:: real2 "-" ureal2 "i")
                 (:: real2 "+i")
                 (:: real2 "-i")
                 (:: "+" ureal2 "i")
                 (:: "-" ureal2 "i")
                 (:: "+i")
                 (:: "-i"))]
  [real2 (:: sign ureal2)]
  [ureal2 (:or uinteger2 (:: uinteger2 "/" uinteger2))]
  [uinteger2 (:: (:+ digit2) (:* #\#))]
  [prefix2 (:or (:: radix2 exactness)
                (:: exactness radix2))]
  [radix2 "#b"]
  [digit2 (:or "0" "1")]
  [num8 (:: prefix8 complex8)]
  [complex8 (:or real8
                 (:: real8 "@" real8)
                 (:: real8 "+" ureal8 "i")
                 (:: real8 "-" ureal8 "i")
                 (:: real8 "+i")
                 (:: real8 "-i")
                 (:: "+" ureal8 "i")
                 (:: "-" ureal8 "i")
                 (:: "+i")
                 (:: "-i"))]
  [real8 (:: sign ureal8)]
  [ureal8 (:or uinteger8 (:: uinteger8 "/" uinteger8))]
  [uinteger8 (:: (:+ digit8) (:* #\#))]
  [prefix8 (:or (:: radix8 exactness)
                (:: exactness radix8))]
  [radix8 "#o"]
  [digit8 (:/ "0" "7")]
  
  [num10 (:: prefix10 complex10)]
  [complex10 (:or real10
                  (:: real10 "@" real10)
                  (:: real10 "+" ureal10 "i")
                  (:: real10 "-" ureal10 "i")
                  (:: real10 "+i")
                  (:: real10 "-i")
                  (:: "+" ureal10 "i")
                  (:: "-" ureal10 "i")
                  (:: "+i")
                  (:: "-i"))]
  [real10 (:: sign ureal10)]
  [ureal10 (:or uinteger10 (:: uinteger10 "/" uinteger10) decimal10)]
  [uinteger10 (:: (:+ digit10) (:* #\#))]
  [prefix10 (:or (:: radix10 exactness)
                 (:: exactness radix10))]
  [radix10 (:? "#d")]
  [digit10 digit]
  [decimal10 (:or (:: uinteger10 suffix)
                  (:: #\. (:+ digit10) (:* #\#) suffix)
                  (:: (:+ digit10) #\. (:* digit10) (:* #\#) suffix)
                  (:: (:+ digit10) (:+ #\#) #\. (:* #\#) suffix))]
  
  [num16 (:: prefix16 complex16)]
  [complex16 (:or real16
                  (:: real16 "@" real16)
                  (:: real16 "+" ureal16 "i")
                  (:: real16 "-" ureal16 "i")
                  (:: real16 "+i")
                  (:: real16 "-i")
                  (:: "+" ureal16 "i")
                  (:: "-" ureal16 "i")
                  "+i"
                  "-i")]
  [real16 (:: sign ureal16)]
  [ureal16 (:or uinteger16 (:: uinteger16 "/" uinteger16))]
  [uinteger16 (:: (:+ digit16) (:* #\#))]
  [prefix16 (:or (:: radix16 exactness)
                 (:: exactness radix16))]
  [radix16 "#x"]
  [digit16 (:or digit (:/ #\a #\f) (:/ #\A #\F))]
  
  
  [suffix (:or "" (:: exponent-marker sign (:+ digit10)))]
  [exponent-marker (:or "e" "s" "f" "d" "l")]
  [sign (:or "" "+" "-")]
  [exactness (:or "" "#i" "#e")])



; --------------------------------------------------------------
; Parser helpers
; --------------------------------------------------------------

(define stx-for-original-property (read-syntax #f (open-input-string "original")))

;; A macro to build the syntax object
(define-syntax (build-so stx)
  (syntax-case stx ()
    ((_ value start end)
     (with-syntax ((start-pos (datum->syntax
                               (syntax end)
                               (string->symbol 
                                (format "$~a-start-pos"
                                        (syntax->datum (syntax start))))))
                   (end-pos (datum->syntax
                             (syntax end)
                             (string->symbol 
                              (format "$~a-end-pos"
                                      (syntax->datum (syntax end))))))
                   (source (datum->syntax
                            (syntax end)
                            'source-name)))
       (syntax
        (datum->syntax
         #f
         value
         (list source 
               (position-line start-pos)
               (position-col start-pos)
               (position-offset start-pos)
               (- (position-offset end-pos)
                  (position-offset start-pos)))
         stx-for-original-property))))))

(define-syntax (syn-error stx)
  (syntax-case stx ()
    ((_ msg token)
     (with-syntax ((source (datum->syntax
                            (syntax token)
                            'source-name)))
       (syntax
        (raise-syntax-error 
         #f 
         msg
         (datum->syntax
          #f
          (token-name (position-token-token token)) ;syntax object "value" field
          (list source 
                (position-line (position-token-start-pos token))
                (position-col (position-token-start-pos token))
                (position-offset (position-token-start-pos token))
                (- (position-offset (position-token-end-pos token))
                   (position-offset (position-token-start-pos token))))
          stx-for-original-property)))))))

;I'm lazy, so I want (begin (define id val) ... expr) to work everywhere
(define-syntax begin/define
  (syntax-rules ()
    [(begin/define (define id expr) forms ...)
     (let ([id expr])
       (begin/define forms ...))]
    [(begin/define expr)
     expr]))
     
      

; --------------------------------------------------------------
; Parser
; --------------------------------------------------------------

#|
Grammar
 expr
   : id
   | constant

   //Application with expr in app position
   | '_(' expr ')' '(' expr_lst ... ')' -> '(' expr expr_lst ... ')'

   //Normal application f(a ...)
   | id '(' expr ... ')' -> '(' id expr ... ')'

   //Infix application
   | '(|' expr_0 expr_1 expr_2 '|)' -> '(' expr_1 expr_0 expr_2 ')'
   | '{'  expr_0 expr_1 expr_2 '}'  -> '(' expr_1 expr_0 expr_2 ')'

   //Normal prefix s-exprs
   | '[' expr ... ']' -> '(' expr ... ')'
   
   //Rules that use colon to replace parens
   | id {no_ws()}? ':' ( { checkIndentation() }? expr ) ... -> '('expr ... ')'
   | ':' ( { checkIndentation() }? expr ) ... -> '(' expr ... ')'
   | id '(' expr_0 ... '):' ( { checkIndentation() }? expr_1 ) ... -> '(' id expr_0 ... expr_1 ...')'
   ;

Rules for checkIndentation()
  - If there is a token on the same line ->
    - Subsequent lines must be indented more than the first char in the COLON, CP_COLON, or ID_COLON token.
    - Keep tokens on the same line (don't end indenting yet)
  - if there isn't another token on the same line ->
    - subsequent lines must be indented more than first char on the line in which the group started
    - Is this needed????  It allows:
      let: :
        [a 1]
        [b 2]
       +(a b)
      which is rather odd behavior...
  - multiple tokens on an included subsequent line are kept
  - multiple subsequent lines must have the same indentation
    - This, btw, combined with the previous case, allows:
    let: :
     [a 1]
     [b 2]
      +(a b)
    which is totally bizarre behavior.

NOTE: indent on subsequent lines must be greater than the "colon". 
However, it uses the start position of the token containing the colon, which is COLON, CO_COLON, or ID_COLON.
That last one allows:
  when: cond()
   display("true")
Since "when:" is the whole token.  This was unintentional, but a good feature.
It also explains why I have a separate rule for min column when no tokens are on the first line after the colon.

|#

(define (alt-exp-ll1 source-name)
  
  ;Internal util functions
  
  ;Only LL uses lex to grab a new token.
  (define next-tok null)
  (define current-line -1)
  (define current-indent 0)
  (define (ll lex)
    ;Get a new token if we don't have one yet
    (when (eq? next-tok null)
      (set! next-tok (lex))
      ;Need to update line / indent values
      (when ((position-line (position-token-start-pos next-tok)) . > . current-line)
        (set! current-line (position-line (position-token-start-pos next-tok)))
        (set! current-indent (position-col (position-token-start-pos next-tok)))))
    ;Return token
    next-tok)
  
  ;Consume a token, making sure it's the expected type
  (define (consume target lex)
    (if (eq? target (token-name (position-token-token (ll lex))))
        (set! next-tok null) ;eat that token
        (syn-error (format "parse failed, expected ~a, got ~a\n" target (token-name (position-token-token next-tok))) next-tok))
    (void))
  
  ;Consume a token, making sure it's the expected type, but also return it
  (define (get-consume target lex)
    (define tok (ll lex))
    (set! next-tok null) ;eat the token
    (unless (eq? target (token-name (position-token-token tok)))
        (syn-error (format "parse failed, expected ~a, got ~a\n" target (token-name (position-token-token tok))) tok))
    tok) ;but also return it
  
  
  ;Need to know what can start an altexp (for reading alt-exp-list, since it knows to read another by checking this)
  (define first-of-altexp '(ID DATUM COP INOP SOP _OP ID_COLON COLON HASHOP QUOTE QUASIQUOTE UNQUOTE UNQUOTE-SPLICING SYNTAX-QUOTE))
    
  ;altexp : ... lots of options
  (define (read-altexp lex)
    (match (token-name (position-token-token (ll lex)))
      ['DATUM
       (begin/define
         (define $1-tok (get-consume 'DATUM lex))
         (define $1 (token-value (position-token-token $1-tok)))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $1-end-pos (position-token-end-pos $1-tok))
         (build-so $1 1 1))]
      ; Application with expr in app position
      ;(_OP altexp CP OP altexp-list CP)
      ['_OP
       (begin/define
         (define $1-tok (get-consume '_OP lex))
         (define $2 (read-altexp lex))
         (define $3_ (consume 'CP lex))
         (define $4_ (consume 'OP lex))
         (define $5 (read-altexp-list lex))
         (define $6-tok (get-consume 'CP lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $6-end-pos (position-token-end-pos $6-tok))
         (build-so (cons $2 $5) 1 6))]
      ; Infix application
      ;(INOP altexp altexp altexp INCP)
      ['INOP 
       (begin/define
         (define $1-tok (get-consume 'INOP lex))
         (define $2 (read-altexp lex))
         (define $3 (read-altexp lex))
         (define $4 (read-altexp lex))
         (define $5-tok (get-consume 'INCP lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $5-end-pos (position-token-end-pos $5-tok))
         (build-so (list $3 $2 $4) 1 5))]
      ; Normal prefix s-exprs
      ;(SOP altexp-list SCP)
      ['SOP
       (begin/define
         (define $1-tok (get-consume 'SOP lex))
         (define $2 (read-altexp-list lex))
         (define $3-tok (get-consume 'SCP lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $3-end-pos (position-token-end-pos $3-tok))
         (build-so $2 1 3))]
      #|;(_OP altexp-list CP)
      ['_OP
       (begin/define
         (define $1-tok (get-consume '_OP lex))
         (define $2 (read-altexp-list lex))
         (define $3-tok (get-consume 'CP lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $3-end-pos (position-token-end-pos $3-tok))
         (build-so $2 1 3))]|#
      ; Other R5Rs Scheme bits of syntax
      ;(HASHOP altexp-list CP)
      ['HASHOP 
       (begin/define
         (define $1-tok (get-consume 'HASHOP lex))
         (define $2 (read-altexp-list lex))
         (define $3-tok (get-consume 'CP lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $3-end-pos (position-token-end-pos $3-tok))
         (build-so (list->vector $2) 1 3))]
      ;(QUOTE altexp)
      ['QUOTE
       (begin/define
         (define $1-tok (get-consume 'QUOTE lex))
         (define $2 (read-altexp lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position $2)
                                           (syntax-line $2)
                                           (syntax-column $2)))
         (build-so (list 'quote $2) 1 2))]
      ;(QUASIQUOTE altexp)
      ['QUASIQUOTE
       (begin/define
         (define $1-tok (get-consume 'QUASIQUOTE lex))
         (define $2 (read-altexp lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position $2)
                                           (syntax-line $2)
                                           (syntax-column $2)))
         (build-so (list 'quasiquote $2) 1 2))]
      ;(UNQUOTE altexp)
      ['UNQUOTE
       (begin/define
         (define $1-tok (get-consume 'UNQUOTE lex))
         (define $2 (read-altexp lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position $2)
                                           (syntax-line $2)
                                           (syntax-column $2)))
         (build-so (list 'unquote $2) 1 2))]
      ;(UNQUOTE-SPLICING altexp)
      ['UNQUOTE-SPLICING
       (begin/define
         (define $1-tok (get-consume 'UNQUOTE-SPLICING lex))
         (define $2 (read-altexp lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position $2)
                                           (syntax-line $2)
                                           (syntax-column $2)))
         (build-so (list 'unquote-splicing $2) 1 2))]
      ; Other Racket symbols
      ;(SYNTAX-QUOTE altexp)
      ['SYNTAX-QUOTE
       (begin/define
         (define $1-tok (get-consume 'SYNTAX-QUOTE lex))
         (define $2 (read-altexp lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position $2)
                                           (syntax-line $2)
                                           (syntax-column $2)))
         (build-so (list 'syntax $2) 1 2))]
      ; Rules that use colon to replace parens
      ;(ID_COLON altexp-indent-list)
      ['ID_COLON
       (begin/define
         (define $1-tok (get-consume 'ID_COLON lex))
         (define $1 (token-value (position-token-token $1-tok)))
         (define $rest (read-altexp-indent-list $1-tok lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position (last $rest))
                                           (syntax-line (last $rest))
                                           (syntax-column (last $rest))))
         (build-so (cons $1 $rest) 1 2))]
      ;(COLON altexp-indent-list)
      ['COLON
       (begin/define
         (define $1-tok (get-consume 'COLON lex))
         (define $rest (read-altexp-indent-list $1-tok lex))
         (define $1-start-pos (position-token-start-pos $1-tok))
         (define $2-end-pos (make-position (syntax-position (last $rest))
                                           (syntax-line (last $rest))
                                           (syntax-column (last $rest))))
         (build-so $rest 1 2))]
      ;Common prefix refactoring on rules with ID ...
      ['ID
       (begin/define
         (define $1-tok (get-consume 'ID lex))
         (if (eq? 'OP (token-name (position-token-token (ll lex))))
             ;ID OP ...
             (read-finish-id-app lex $1-tok)
             ;Just plain ID
             (begin/define
               (define $1 (token-value (position-token-token $1-tok)))
               (define $1-start-pos (position-token-start-pos $1-tok))
               (define $1-end-pos (position-token-end-pos $1-tok))
               (build-so $1 1 1))))]
      [else 
       (begin
         (syn-error (format "parse error while reading altexp, unexpected starting token ~a\n" 
                 (token-name (position-token-token (ll lex)))) (ll lex))
         (set! next-tok null) ;consume that token (we alread read one in order to be here, now clear it to force new read)
         null)]))
  
  ;The rest of the common-prefix ID rule, when ID is followed by OP (an application)
  (define (read-finish-id-app lex $1-tok)
    (define $1 (token-value (position-token-token $1-tok)))
    (define $2_ (consume 'OP lex))
    (define $3 (read-altexp-list lex))
    (define $1-start-pos (position-token-start-pos $1-tok))
    (match (token-name (position-token-token (ll lex)))
      ; Normal application f(a ...)
      ;(ID OP altexp-list CP)
      ['CP
       (begin/define
         (define $4-tok (get-consume 'CP lex))
         (define $4-end-pos (position-token-end-pos $4-tok))
         (build-so (cons $1 $3) 1 4))]
      ;(ID OP altexpr-list CPCOLON altexp-indent-list)
      ['CPCOLON
       (begin/define
         (define $4-tok (get-consume 'CPCOLON lex))
         (define $rest (read-altexp-indent-list $4-tok lex))
         (define $5-end-pos (if (empty? $rest)
                                (syn-error (format "Expr of the form 'id( expr ... ): {{expr ...}}' missing the rest of exprs\n")
                                           $4-tok)
                                (make-position (syntax-position (last $rest))
                                           (syntax-line (last $rest))
                                           (syntax-column (last $rest)))))
         (build-so (cons $1 (append $3 $rest)) 1 5))]
      [else
       (begin
         (syn-error (format "parse error while reading rest of id-app, unexpected token ~a\n" 
                 (token-name (position-token-token (ll lex)))) (ll lex))
         null)]))
  
 
  
  ;altexpr-list : mt | alt-exp altexp-list ;
  (define (read-altexp-list lex)
    (if (member (token-name (position-token-token (ll lex))) first-of-altexp)
        (let ([$1 (read-altexp lex)]
              [$2 (read-altexp-list lex)])
          (cons $1 $2)) ;NOTE: the list is in the right order
        null)) ;null if no more exprs to read
  
  ;This reads in ({checkIndent()}? altexp)* using column info from colon-tok
  ;to check the indentation.  We also need to know indentation of the line on
  ;which colon-tok resides, so we have to cheat and ask the lexer that.
  (define (read-altexp-indent-list colon-tok lex)
    (define line (position-line (position-token-start-pos colon-tok)))
    (define line-indent current-indent)
    (define column (position-col (position-token-start-pos colon-tok)))
    (define tok-on-line? #f)
    (define indent -1)
    (define last-line current-line)
    
    (define (indent-done?)
      (define tok-line (position-line (position-token-start-pos (ll lex))))
      (define tok-col (position-col (position-token-start-pos (ll lex))))
      
      (cond
        [((token-name (position-token-token (ll lex))) . eq? . 'EOF)
         #t] ;EOF ends it, of course
        [(tok-line . equal? . line)
         (begin
           (set! tok-on-line? #t) ;subsequent lines need indent > colon column
           #f)]
        [(tok-line . equal? . last-line)
         #f] ;still on the same line as a previously included token
        [(and (tok-line . > . last-line)
              (tok-col . > . (if tok-on-line? column line-indent)) ;valid indentation
              (or (indent . equal? . -1) ;setting indent group, or matches indent group
                  (indent . equal? . tok-col)))
         (begin
           (when (indent . equal? . -1)
             (set! indent tok-col)) ;this is now the indent we use
           (set! last-line tok-line) ;remember line
           #f)]
        [else ;not on same line as previous token, and it either has invalid indent or doesn't match indent group
         #t]))
    
    (let loop ([lst '()])
      (if (or (not (member (token-name (position-token-token (ll lex))) first-of-altexp))
              (indent-done?))
          lst ;next token doesn't fit in this group, so done
          (loop (append lst (list (read-altexp lex))))))) ;keep appending altexprs
  
   ;s : altexp-list ;
  (define (read-s lex)
    (define $1 (read-altexp-list lex))
    $1)
  
  ;Returns a function to read files given lexer
  (lambda (lex)
    (read-s lex)));end parser

  
(define (read-syntax-helper source-name in-port)
  (port-count-lines! in-port)
  ((alt-exp-ll1 source-name) (lambda () (alt-exp-lexer in-port))))

;Read some syntax, keeping stx info
(define my-read-syntax
  (case-lambda 
    [(source-name) 
     (read-syntax-helper source-name (current-input-port))]
    [(source-name in-port) 
     (read-syntax-helper source-name in-port)]))

;Read without syntax info (so just strip it)
(define (my-read in)
  (syntax->datum (read-syntax #f in)))

;Functions required to be a reader -- can't use the name "read-syntax" because we use Rkt's reader up above.
(provide (rename-out [my-read read]
                     [my-read-syntax read-syntax]))

