#lang alt-exp


#|
Spec: define a case construct syntactically just like that of Racket.
In terms of semantics:

- each branch automatically falls through to the next,

- the last one returns its answer since it has no next clause, and

- any branch can contain (break <expr>), which evaluates <expr> and
returns its value as that of the entire case.

In honor of its behavor, we'll call this cas-cad-e.  Thus,

(define (cas1 v)
   (cas-cad-e v
            ((1) (display "1"))
            ((2) (display "2") (break 2)
            ((3) 3))))

(cas1 1) ==> 2       (and prints "12")
(cas1 2) ==> 2       (and prints "2")
(cas1 3) ==> 3       (and prints nothing)
(cas1 4) ==> <void>  (and prints nothing)

|#

;Expand last clause first, building the "next" function and the "ifs" computation as we go.
;Each action calls the "next" action, unless a break is issued.
define-syntax: cas-cad-e-inner(stx)
  syntax-case(stx [break]):
    :[_ [next ifs] e] ;base case, all has been converted to the ifs
     #'ifs
    :[_ [next ifs] e [c s ...] ... [[cls] stm ... break(v)]]
     #'let: :[action [lambda [x] stm ... v]] ;have a break, so action returns the value
         cas-cad-e-inner:
           [action if(equal?(e cls) action(0) ifs)]
           e
           [c s ...] ...
    :[_ [next ifs] e [c s ...] ... [[cls] stm ... stm_v]]
     #'let: :[action [lambda [x] stm ... next(stm_v)]] ;no break, so pass along ending value in case we're last.
         cas-cad-e-inner:
           [action if(equal?(e cls) action(0) ifs)]
           e
           [c s ...] ...

;Just launches inner with a "next" that returns the last value (if we get that far w/o a break).
; For this reason all "next" functions are passed the last value in a clause's statements, though
; it is ignored by most.
define-syntax: cas-cad-e(stx)
  syntax-case(stx [break]):
    :[_ e [c s ...] ...]
     syntax:
       cas-cad-e-inner:
         [[lambda [x] x] void()] ;the "next" function and the remainder of the "ifs" portion of computation.
         e
         [c s ...] ...
    

;;; TEST

define: cas1(v)
  cas-cad-e(v):
    :[1] display("1")
    :[2] display("2") break(2)
    :[3] 3
    
cas1(1) ;==> 2 and prints "12"

cas1(2) ;==> 2 and prints "2"

cas1(3) ;==> 3 and prints nothing

cas1(4) ;==> void and prints nothing
